import boto3
import json
import logging
import sys
import os
import base64
import datetime
import time
from boto3.dynamodb.conditions import Key, Attr

#---------------------------------------------------------------------------------
# Global Scope constants (reused across function invocations).
#---------------------------------------------------------------------------------
log = logging.getLogger(__name__)
thisAcctBotoSession = boto3.session.Session()

# Domain join user with minimum privileges: https://docs.aws.amazon.com/directoryservice/latest/admin-guide/directory_join_privileges.html
# Assumptions:
# AWS Managed Active directory is deployed in a central security master account. in a core region (say us-east-1).
# All VPCs are connected to the VPC hosting AWS Managed Active directory in the central security master account.
# This Lambda is deployed in the central security master account core region.
# DynamoDB table with account info is deployed in the central security master account core region.
# A EC2 with ADTools is deployed in the security master account core region for this Lambda to invoke deletion of AD Computer objects.

#---------------------------------------------------------------------------------
# Join instance to AWS MAD.
# C:\ProgramData\Amazon\SSM\InstanceData\<instance-id>\document\orchestration\<execution-id>\awsrunPowerShellScript\0.awsrunPowerShellScript\_script.ps1
# /var/lib/amazon/ssm/<instance-id>/document/orchestration/<execution-id>/awsrunShellScript/0.awsrunShellScript/_script.sh
#---------------------------------------------------------------------------------
def joinAD(msgBody, acctInfo, instanceInfo, ssmClient):
    platform = 'unknown'
    if 'Platform' in instanceInfo:
        platform = instanceInfo['Platform']
    log.debug('Instance platform: ' + platform)
    if (platform == "windows"):
        scriptFile = "joinDomain.ps1"
        ssmRunCommand = "AWS-RunPowerShellScript"
    else:
        scriptFile = "joinDomain.sh"
        ssmRunCommand = "AWS-RunShellScript"
    fin = open(scriptFile, "rt")
    commands = fin.read().splitlines()
    cmdStr = 'joinAD "%s" "%s" "%s" "%s" "%s"'
    cmdStr = cmdStr % (acctInfo['dnsName'],acctInfo['dnsIP1'],acctInfo['dnsIP2'],acctInfo['adJoinUserName'],getADJoinPasswd(acctInfo))
    commands.append(cmdStr)
    params = { 'commands': commands }
    sendCommandResponse = ssmClient.send_command(InstanceIds=[instanceInfo['InstanceId']], DocumentName=ssmRunCommand, Parameters=params, CloudWatchOutputConfig={'CloudWatchOutputEnabled': True})
    if 'Command' in sendCommandResponse and 'CommandId' in sendCommandResponse['Command']:
        log.info(sendCommandResponse['Command']['CommandId'])

#---------------------------------------------------------------------------------
# Get ssmClient:
#---------------------------------------------------------------------------------
def getSsmClient(instanceInfo, acctInfo, otherAcctBotoSession, msgBody):
    if otherAcctBotoSession: # Request not from this account
        log.debug('Found otherAcctBotoSession.')
        ssmClient = otherAcctBotoSession.client('ssm', region_name=msgBody['region'])
    else:
        log.debug('Using thisAcctBotoSession')
        ssmClient = thisAcctBotoSession.client('ssm', region_name=msgBody['region'])
    return ssmClient

#---------------------------------------------------------------------------------
# Retrieve account info from DynamoDB.
#---------------------------------------------------------------------------------
def getAcctInfo(requestAcctNum):
    try:
        dynamodb = thisAcctBotoSession.resource('dynamodb')
        log.debug(os.environ['dynamoDBTableName'])
        table = dynamodb.Table(os.environ['dynamoDBTableName'])
        response = table.query(KeyConditionExpression=Key('accountId').eq(requestAcctNum))
        return response['Items'][0]
    except:
        log.error('MAD-AD-933: Error while retrieving account metadata information')
        return None

#---------------------------------------------------------------------------------
# Generate Shell script for Linus instances.
#---------------------------------------------------------------------------------
def getEC2TagValue(instanceInfo, tagKey):
    if 'Tags' not in instanceInfo:
        return ''
    for tag in instanceInfo['Tags']:
        if (tag['Key'].lower() == tagKey.lower()):
            return tag['Value']
    return ''

#---------------------------------------------------------------------------------
# Assume role to setup remote Boto session.
#---------------------------------------------------------------
def getOtherAcctBotoSession(acctInfo, msgRegion, myIdentity):
    try:
        stsClient = thisAcctBotoSession.client('sts')
        creds=stsClient.assume_role(RoleArn=acctInfo['assumeRoleARN'], RoleSessionName=myIdentity['Account']+'-adJoinLambda')
        accessKey =creds['Credentials']['AccessKeyId']
        accessSecret = creds['Credentials']['SecretAccessKey']
        sessionToken = creds['Credentials']['SessionToken']
        otherAcctBotoSession = boto3.session.Session(aws_access_key_id=accessKey, aws_secret_access_key=accessSecret, aws_session_token=sessionToken, region_name=msgRegion)
        return otherAcctBotoSession
    except:
        log.error('MAD-AD-934: Error while assuming role')
        return None

#---------------------------------------------------------------------------------
#  Retrieve instance info local or remote.
#---------------------------------------------------------------
def getInstanceInfo(msgBody, otherAcctBotoSession):
    instanceId = msgBody['detail']['instance-id']
    if otherAcctBotoSession: # Request not from this account
        ec2Client = otherAcctBotoSession.client('ec2', region_name=msgBody['region'])
    else:
        ec2Client = thisAcctBotoSession.client('ec2', region_name=msgBody['region'])
    try:
        ec2Response = ec2Client.describe_instances(InstanceIds=[instanceId])
        return ec2Response['Reservations'][0]['Instances'][0]
    except:
        log.error('MAD-AD-935: Error while retrieving instance information')
        return None

#---------------------------------------------------------------------------------
# Handle removal of AD Computer object on instance termination.
#---------------------------------------------------------------------------------
def handleInstanceTermination(instanceInfo, acctInfo, msgBody, ssmClient):
    computerName = getEC2TagValue(instanceInfo, 'AD-ComputerName')
    if not computerName:
        log.error('MAD-AD-936: AD-ComputerName tag not found on instance during termination: ' + str(msgBody))
        return
    # ssmClient = thisAcctBotoSession.client('ssm')
    fin = open("leaveDomain.ps1", "rt")
    commands = fin.read().splitlines()
    cmdStr = 'leaveAD "%s" "%s" "%s" "%s"'
    cmdStr = cmdStr % (acctInfo['dnsName'],acctInfo['adJoinUserName'],getADJoinPasswd(acctInfo),computerName)
    commands.append(cmdStr)
    params = { 'commands': commands }
    targets = []
    targets.append({'Key': 'tag:'+os.environ['ADToolsHostTagKey'], 'Values': [os.environ['ADToolsHostTagValue']]})
    try:
        sendCommandResponse = ssmClient.send_command(DocumentName='AWS-RunPowerShellScript', Parameters=params, Targets=targets, CloudWatchOutputConfig={'CloudWatchOutputEnabled': True})
        log.debug(sendCommandResponse)
    except Exception as e:
        log.error('MAD-AD-937: **** Caught exception trying to invoke SSM AWS-RunPowerShellScript Command: ' + str(e))

#---------------------------------------------------------------------------------
# Retrieve decrypted AD Join password from account info and SSM Parameter store.
#---------------------------------------------------------------------------------
def getADJoinPasswd(acctInfo):
    ssmClient = thisAcctBotoSession.client('ssm')
    ssmResp = ssmClient.get_parameter(Name=acctInfo['adJoinUserPwParam'], WithDecryption=True)
    passwd = ssmResp['Parameter']['Value']
    b64Passwd = base64.b64encode(passwd.encode('utf-8'))
    return b64Passwd.decode('utf-8')

#---------------------------------------------------------------------------------
# Main handler. Retrieve SQS message attributes and environment variables.
# Identify platform of the instance to invoke appropriate AD join function.
#---------------------------------------------------------------------------------
def lambda_handler(event, context):
    logLevel=os.environ['LogLevel']
    if not logLevel:
        log.setLevel(20)
    else:
        log.setLevel(int(logLevel))

    log.debug(event)
    log.info(json.dumps(event))

    try:
        msgBody = json.loads(event['Records'][0]['body'])
        log.info(msgBody)
    except Exception as err:
        log.error('Error trying to retrieve message body: %s', str(err))
        return

    # Retrieve account info from DynamoDB for the account listed on CloudWatch Instance state event SQS message
    acctInfo = getAcctInfo(msgBody['account'])
    log.debug(acctInfo)
    if not acctInfo:
        log.error('MAD-AD-938: **** Account not found in DyanamoDB: ' + msgBody['account'])
        return

    stsClient = thisAcctBotoSession.client('sts')
    myIdentity = stsClient.get_caller_identity()
    otherAcctBotoSession = None
    requestFromThisAccount = (myIdentity['Account'] == msgBody['account'])
    # if event/request originated from different account then setup boto session using STS
    if not requestFromThisAccount:
        log.debug("Request received from different account.")
        otherAcctBotoSession = getOtherAcctBotoSession(acctInfo, msgBody['region'], myIdentity)

    # retrieve instance info, use STS session to cross account if needed.
    instanceInfo = getInstanceInfo(msgBody, otherAcctBotoSession)
    if not instanceInfo:
        log.error('MAD-AD-941: **** Unable to retrieve instance info.')
        return
    log.info(instanceInfo)

    # check if instance is in running or terminated state, else return.
    if not (instanceInfo['State']['Name'] == 'running' or instanceInfo['State']['Name'] == 'terminated'):
        log.error('MAD-AD-942: ***** Instance not in running or terminated state.')
        return

    ssmClient = getSsmClient(instanceInfo, acctInfo, otherAcctBotoSession, msgBody)

    log.debug('Check for termination of instance')
    # if instance has been terminated then handle termination - cleanup AD Computer object.
    if (instanceInfo['State']['Name'] == 'terminated'):
        handleInstanceTermination(instanceInfo, acctInfo, msgBody, ssmClient)
        return

    computerNameTagVal = getEC2TagValue(instanceInfo, 'AD-ComputerName')
    if computerNameTagVal:
        log.info("Instance %s alredy joined domain, has computer name tag: %s",instanceInfo['InstanceId'],computerNameTagVal)
        return

    #check autoadjoin flag set by workstream instance
    autoadjoinTagVal = getEC2TagValue(instanceInfo, 'autoadjoin')
    if autoadjoinTagVal and autoadjoinTagVal.lower() == 'false':
        log.info('autoadjoinTagVal is false, not doing ADJoin')
        return
    
    joinAD(msgBody, acctInfo, instanceInfo, ssmClient)

