Function joinAD([string]$directoryDNSName, [string]$dnsIP1, [string]$dnsIP2, [string]$domainJoinUser, [string]$domainJoinUserPW) {
    try {
        # Setup Application event log for AD Join.
        If (-Not [System.Diagnostics.EventLog]::SourceExists("adJoinScript")) {
            New-EventLog -LogName Application -Source adJoinScript
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 100 -Message 'adJoinScript event log created.'
        }

        # Retrieve Instance info from Metadata, retrieve accountId, region and instanceId info.
        $instanceInfo = Invoke-WebRequest 'http://169.254.169.254/latest/dynamic/instance-identity/document/' -UseBasicParsing | ConvertFrom-Json
        $accountId = $instanceInfo.accountId
        $awsRegion = $instanceInfo.region
        $instanceId = $instanceInfo.instanceId
        If (-Not ($accountId -And $awsRegion -And $instanceId) ) {
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Error -EventID 110 -Message 'Unable to retrieve accountId / awsRegion / instnaceId from instance metadata.'
            exit -1
        }

        # If function was not invoked with parameter values, then try to load those from 'adjoin' SSM parameter.
        If (-Not ($directoryDNSName -And $dnsIP1 -And $dnsIP2 -And $domainJoinUser -And $domainJoinUserPW)) {
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 120 -Message 'Retrieving AD Join parameters from SSM.'
            $adJoinParamValue = (Get-SSMParameterValue -Name 'adjoin' -WithDecryption $true).Parameters.Value
            If ($adJoinParamValue) {
                $adJoinParams = $adJoinParamValue.Split(',')
                If ($adJoinParams.Length -ne 5) {
                    Write-EventLog -LogName Application -Source adJoinScript -EntryType Error -EventID 121 -Message 'Unable to derive AD Join parameters from SSM Parameter.'
                    exit -1
                }
                $directoryDNSName=$adjoinParams[0]
                $dnsIP1=$adjoinParams[1]
                $dnsIP2=$adjoinParams[2]
                $domainJoinUser=$adjoinParams[3]
                $domainJoinUserPW=$adjoinParams[4]
            }
        }
        # If we are not give necessary parameter values or unable to retrieve from 'adjoin' SSM parameter, then exit.
        If (-Not ($directoryDNSName -And $dnsIP1 -And $dnsIP2 -And $domainJoinUser -And $domainJoinUserPW)) {
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Error -EventID 130 -Message 'Unable to retrieve directoryDNSName / dnsIP1 / dnsIP2 / domainJoinUser / domainJoineUserPW.'
            exit -1
        }

        # Test connectivity with Domain Controller IP1 and IP2.
        If (-Not ((Test-Connection -ComputerName $dnsIP1 -Count 1 -Quiet) -And (Test-Connection -ComputerName $dnsIP2 -Count 1 -Quiet)) ) {
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Error -EventID 140 -Message 'Failed to test connection with DC1 $dnsIP1 / DC2 $dnsIP2.'
            exit -1
        }

        # Retrieve instance Tags and look for 'group' and 'bastion' tag values. If neither is found then exit with error.
        # If 'bastion' tag is found then use it to derive AD group name, else use 'group' tag.
        $tags = Get-EC2Tag -Region $instanceInfo.region -Filter @{Name = "resource-id"; Values = $instanceInfo.instanceId }
        If (-Not $tags) {
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Error -EventID 111 -Message 'Unable to retrieve instance tags.'
            exit -1
        }
        $adGroupName = $tags.Where( { $_.Key -eq "adGroupName" }).Value
        $directoryNetBiosName = $directoryDNSName.split('.')[0]
        Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 200 -Message "AD User Group: $adGroupName."

        # Do AD Join. If success, then add 'AD-ComputerName' tag with the value of hostname - that matches computer object under AD.
        # This tag value is used to cleanup AD computer object when the instance terminates.
        # Install RDS-RD-Server to support multiple RDP sessions.
        If ((gwmi win32_computersystem).partofdomain -ne $true) {
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 170 -Message "Joining Computer to domain: $directoryDNSName."
            $Password = [System.Text.Encoding]::UTF8.GetString([System.Convert]::FromBase64String($domainJoinUserPW)) | ConvertTo-SecureString -asPlainText -Force
            $JoinCredential = New-Object System.Management.Automation.PSCredential("$domainJoinUser@$directoryDNSName", $Password)
            Add-Computer -DomainName "$directoryDNSName" -Credential $JoinCredential
            If ($?) {
                Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 173 -Message "Add-Computer to $directoryDNSName Success."
            }
            Else {
                $errMsg = $Error[0].Exception.Message
                Write-EventLog --LogName Application -Source "adJoinScript" -EntryType Error -EventID 172 -Message "Failed to join domain. error: $errMsg"
                exit -1
            }
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 171 -Message "Adding RD TL components."
            Add-WindowsFeature -Name RDS-RD-Server -IncludeAllSubFeature
            Set-ItemProperty "HKLM:\Software\Microsoft\Windows\CurrentVersion\RunOnce" -Name '!RegisterRDL' -Value "c:\WINDOWS\system32\WindowsPowerShell\v1.0\powershell.exe -noexit -command '$ServerFQDN=(Get-WmiObject win32_computersystem).DNSHostName+"."+(Get-WmiObject win32_computersystem).Domain;New-SessionDeployment -ConnectionBroker $ServerFQDN -WebAccessServer $ServerFQDN -SessionHost $ServerFQDN'"
            $autoadjoinTag = $tags.Where( { ($_.Key).ToLower() -eq "autoadjoin" }).Value
            # If 'autoadjoin' tag with 'false' value is found then do not reboot instance. Instance owner will be responsible to reboot the instance.
            If (-Not ($autoadjoinTag -And $autoadjoinTag.ToLower() -eq "false")) {
                Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 174 -Message "Rebooting to complete domain join."
                exit 3010
            }
            Else {
                Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 175 -Message "Please reboot to complete domain join."
                exit 0
            }
        }
        Else {
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 180 -Message "Computer already joined to domain: $directoryDNSName."
        }

        # Add AD group to Remote Desktop Users group. This will not work before instance is reboot. Refer above for 'autoadjoin' tag usage.
        $members = Get-LocalGroupMember -Name "Remote Desktop Users" | Select -ExpandProperty Name
        If ($members -contains $directoryNetBiosName+'\'+$ADGroupName.trim()) {
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 210 -Message "$adGroupName already present in Remote Desktop Users group."
        }
        Else {
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 220 -Message "Adding $adGroupName to Remote Desktop Users group."
            Add-LocalGroupMember -Group "Remote Desktop Users" -Member $ADGroupName
            If (-Not $?) {
                $errMsg = $Error[0].Exception.Message
                Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 221 -Message "Error Adding $adGroupName to Remote Desktop Users group: $errMsg"
            }
        }
        
        # Add AD group to Administrators group. This will not work before instance is reboot. Refer above for 'autoadjoin' tag usage.
        $members = Get-LocalGroupMember -Name "Administrators" | Select -ExpandProperty Name
        If ($members -contains $directoryNetBiosName+'\'+$ADGroupName.trim()) {
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 230 -Message "$adGroupName already present in Administrators Users group."
        }
        Else {
            Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 240 -Message "Adding $adGroupName to Administrators Users group."
            Add-LocalGroupMember -Group "Administrators" -Member $ADGroupName
            If (-Not $?) {
                $errMsg = $Error[0].Exception.Message
                Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 241 -Message "Error Adding $adGroupName to Administrators group: $errMsg"
            }
        }

        Write-EventLog -LogName Application -Source adJoinScript -EntryType Information -EventID 260 -Message "Adding AD-ComputerName: $computerName tag."
        $tag = New-Object Amazon.EC2.Model.Tag
        $tag.Key = "AD-ComputerName"
        $tag.Value = hostname
        New-EC2Tag -Resource $instanceInfo.instanceId -Tag $tag   

    }
    catch [Exception]  {
        $msg = $_.Exception.Message
        Write-Error ("MAD-AD-900: " + $msg)
        exit -1
    }
}
