#!/bin/bash

function joinAD
{
    # Identify if this is 'ubuntu' OS.
    if [[ -f /etc/os-release ]]
    then
        OS_ID=$(grep "^ID=" /etc/os-release | sed "s/^ID=//; s/\"//g")
    fi
    logger "ADJoin: OS_ID: ${OS_ID}"
  
    # Install necessary packages, if not already done in the base AMI.
    if [[ ${OS_ID} == "ubuntu" ]]
    then
        # apt-get update
        # apt-get -y upgrade
        # Expect above to be done in the AMI of the launced instance.
        export DEBIAN_FRONTEND=noninteractive
        apt-get -y install sssd realmd krb5-user samba-common packagekit adcli unzip
    else
        # yum -y update
        # Expect above to be done in the AMI of the launced instance.
        yum -y install sssd realmd krb5-workstation samba-common-tools adcli unzip
    fi
    if [[ $? -ne 0 ]]
    then
        logger "ADJoin: install failed."
        exit 1
    fi

    # setup aws cli
    awspath=$(which aws)
    if [[ $? -ne 0 ]]
    then
        if [[ ! -f "/usr/local/bin/aws" ]]
        then
            curl "https://s3.amazonaws.com/aws-cli/awscli-bundle.zip" -o "/tmp/awscli-bundle.zip"
            unzip /tmp/awscli-bundle.zip -d /tmp
            /tmp/awscli-bundle/install -i /usr/local/aws -b /usr/local/bin/aws
        fi
        awsbin="/usr/local/bin/aws"
    else
        awsbin=$awspath
    fi

    # Retrieve accountId, region and instaneId from instance metadata.
    accountId=$(curl -s http://169.254.169.254/latest/dynamic/instance-identity/document/ | grep accountId | sed "s/[ ,\"]//g; s/accountId://g")
    awsRegion=$(curl -s http://169.254.169.254/latest/dynamic/instance-identity/document/ | grep region | sed "s/[ \",]//g; s/^region://")
    instanceId=$(curl -s http://169.254.169.254/latest/dynamic/instance-identity/document/ | grep instanceId | sed "s/[ \",]//g; s/^instanceId://")
    if [[ -z ${accountId} || -z ${awsRegion} || -z ${instanceId} ]]
    then
        logger "ADJoin: Unable to derive accountId / awsRegion / instanceId from Instance Metadata."
        exit 1
    fi

    # Retrieve function parameter values. If not passed in, then retrieve those values from 'adjoin' SSM parameter.
    directoryDNSName="${1}"
    dnsIP1="${2}"
    dnsIP2="${3}"
    domainJoinUser="${4}"
    domainJoinUserPW="${5}"
    if [[ -z ${directoryDNSName} || -z ${dnsIP1} || -z ${dnsIP2} || -z ${domainJoinUser} || -z ${domainJoinUserPW} ]]
    then
        logger "ADJoin: Retrieving AD Join parameters from SSM."
        IFS=',' read -r -a adjoinParams <<< $($awsbin --region ${awsRegion} ssm  get-parameters --name 'adjoin' --with-decryption --query Parameters[0].Value --output text)
        #check for array length
        if [[ ${#adjoinParams[@]} -ne 5 ]]
        then
            logger "ADJoin: “Unable to derive AD Join parameters from SSM Parameter"
            exit 1
        fi
        directoryDNSName=${adjoinParams[0]}
        dnsIP1=${adjoinParams[1]}
        dnsIP2=${adjoinParams[2]}
        domainJoinUser=${adjoinParams[3]}
        domainJoinUserPW=${adjoinParams[4]}
    fi
    # If no required parameter values passed in or from 'adjoin' SSM parameter, then exit with error.
    if [[ -z ${directoryDNSName} || -z ${dnsIP1} || -z ${dnsIP2} || -z ${domainJoinUser} || -z ${domainJoinUserPW} ]]
    then
        logger "ADJoin: Unable to derive directoryDNSName / dnsIP1 / dnsIP2 / domainJoinUser / domainJoinUserPW from adjoin SSM parameter."
        exit 1
    fi

    # If already domain joined then validate DNS entry and set ComuterName tag and exit.
    if [[ -n "$(/usr/sbin/realm list --name-only)" ]]
    then
        logger "ADJoin: Domain $(/usr/sbin/realm list --name-only) already joined"
        $awsbin --region "${awsRegion}" ec2 create-tags --resources "${instanceId}" --tags Key=AD-ComputerName,Value="$(hostname)"
        if [[ $? -ne 0 ]]
        then
            logger "ADJoin: MAD-AD-904: Failed to create AD-ComputerName Tag"
            exit 1
        fi
        exit 0
    fi

    # Validate connectivity with AD domain controller 1 and 2.
    ping -q -c 1 -W 1 "${dnsIP1}"
    dns1PingStatus=$?
    ping -q -c 1 -W 1 "${dnsIP2}"
    dns2PingStatus=$?
    if [[ $dns1PingStatus -eq 1 || $dns2PingStatus -eq 1 ]]
    then
      logger "ADJoin: Unable to ping dns1 or dns2"
      exit 1
    fi

    # Validate directory DNS name pattern.
    if [[ ! ${directoryDNSName} =~ ^([a-z0-9]+(-[a-z0-9]+)*\.)+[a-z]{2,}$ ]]
    then
        logger "ADJoin: Directory DNS pattern error: ${directoryDNSName}"
        exit 1
    fi

    # Validate DNS resolution.
    domainLookup=$(nslookup "${directoryDNSName}" | grep "server can.t find")
    if [[ -n "${domainLookup}" ]]
    then
        logger "ADJoin: Could not resolve DNS ${directoryDNSName}"
    else
        logger "ADJoin: Resolved DNS ${directoryDNSName}"
    fi

    # set hostname (must be < 16 characters, to do proper cleanup when instance terminates)
    if [[ ! ${hostname}  =~ ^[iI][pP]-([0-9a-zA-Z]){8}$ ]]
    then
        host_ip1=$(curl http://169.254.169.254/latest/meta-data/local-ipv4/)
        if [[ ! "$host_ip1" =~ ^([0-9]{1,3}.?){4}$ ]]
        then
            logger "ADJoin: Unable to derive host IP from http://169.254.169.254/latest/meta-data/local-ipv4/"
        else
            logger "ADJoin: Host IP: $host_ip1"
            hostname_ipHex=ip-$(printf '%02x' $(echo $host_ip1 | sed "s/\./ /g"))
            logger "ADJoin: Converted Host IP: $hostname_ipHex"
            if [[ $hostname_ipHex ]]
            then
                logger "ADJoin: Setting hostname to: ${hostname_ipHex}"
                command -v hostnamectl
                if [[ $? -eq 0 ]]
                then
                    logger "hostnamectl found"
                    hostnamectl set-hostname ${hostname_ipHex}
                elif [[ -f /etc/sysconfig/network ]]
                then
                    logger "Updating /etc/sysconfig/network file with hostname."
                    cp /etc/sysconfig/network /etc/sysconfig/network.orig
                    sed -i "s/^HOSTNAME=localhost.localdomain/HOSTNAME=${hostname_ipHex}/" /etc/sysconfig/network
                    echo "" >> /etc/sysconfig/network
                else
                    logger "Unable to update hostname via hostnamectl or /etc/sysconfig/network"
                fi
                # echo 127.0.0.1 ${hostname_ipHex}.${directoryDNSName} ${hostname_ipHex} > /etc/hosts
            fi
        fi
    fi 

    if [[ ${OS_ID} == "ubuntu" ]]
    then
        #disable reverse dns 
        cp /etc/krb5.conf /etc/krb5.conf.orig
        sed -iu  "3irdns = false" /etc/krb5.conf
    fi

    # Retrieve 'bastion' and 'group' tag values.
    directoryNetBiosName=$(echo ${directoryDNSName} | sed "s/\..*$//g")
    adGroupName=$($awsbin --region ${awsRegion} ec2 describe-tags --filters "Name=resource-type,Values=instance" "Name=key,Values=adGroupName" "Name=resource-id,Values=${instanceId}" --query "Tags[].Value" --output text)
    logger "ADJoin: AD Group Name: ${adGroupName}"
    # Generate AD Group DN.
    rootOUPath=$(echo "OU=${directoryNetBiosName}")
    IFS='.' read -r -a dnsArray <<< "${directoryDNSName}"
    for element in "${dnsArray[@]}"
    do
        rootOUPath=$(echo "${rootOUPath},DC=${element}")
    done
    adGroupDN=$(echo "CN=${adGroupName},${rootOUPath}")
    logger "ADJoin: AD Group DN: ${adGroupDN}"

    # Do AD join. If successful then add 'AD-ComputerName' tag with the hostname as value. This will be used when instance terminates to cleanup AD computer object.
    logger "ADJoin: Joining Domain ${directoryDNSName}"
    echo ${domainJoinUserPW} | base64 --decode | /usr/sbin/realm join --user=${domainJoinUser} ${directoryDNSName} --verbose
    if [[ $? -eq 0 ]]
    then
        logger "ADJoin: Joined Domain ${directoryDNSName}"
    else
        logger "ADJoin: Failed to join Domain"
        exit 1
    fi

    # Setup sshd config.
    cp /etc/ssh/sshd_config /etc/ssh/sshd_config.orig
    sed -i "s/^PasswordAuthentication no/#PasswordAuthentication no/" /etc/ssh/sshd_config
    sed -i "s/^#PasswordAuthentication yes/PasswordAuthentication yes/" /etc/ssh/sshd_config
    sed -i "s/^#LogLevel INFO/LogLevel INFO/" /etc/ssh/sshd_config
    sed -i "s/^#MaxAuthTries .*/MaxAuthTries 4/" /etc/ssh/sshd_config
    sed -i "s/^#IgnoreRhosts yes/IgnoreRhosts yes/" /etc/ssh/sshd_config
    sed -i "s/^#HostbasedAuthentication no/HostbasedAuthentication no/" /etc/ssh/sshd_config
    sed -i "s/^#PermitRootLogin .*/PermitRootLogin no/" /etc/ssh/sshd_config
    sed -i "s/^#PermitEmptyPasswords no/PermitEmptyPasswords no/" /etc/ssh/sshd_config
    sed -i "s/^#PermitUserEnvironment no/PermitUserEnvironment no/" /etc/ssh/sshd_config
    sed -i "s/^#ClientAliveInterval .*/ClientAliveInterval 300/" /etc/ssh/sshd_config
    sed -i "s/^#LoginGraceTime .*/LoginGraceTime 60/" /etc/ssh/sshd_config
    sed -i "s/^#Banner .*/Banner \/etc\/issue.net/" /etc/ssh/sshd_config
    echo "" >> /etc/ssh/sshd_config
    echo "AllowUsers * " >> /etc/ssh/sshd_config
    echo "Protocol 2" >> /etc/ssh/sshd_config
    echo "Ciphers aes256-ctr,aes192-ctr,aes128-ctr" >> /etc/ssh/sshd_config
    echo "MACs hmac-sha2-512,hmac-sha2-256" >> /etc/ssh/sshd_config

    # Setup sssd config.
    logger "ADJoin: Adding AD Group to SSSD: ${adGroupDN}"
    grep -v "^ad_access_filter " /etc/sssd/sssd.conf > /etc/sssd/sssd.conf.tmp
    echo "ad_access_filter = (memberOf=${adGroupDN})" >> /etc/sssd/sssd.conf.tmp
    echo "debug_level = 5" >> /etc/sssd/sssd.conf.tmp

    if [[ ${OS_ID} == "ubuntu" ]]
    then
        sed -i "s/^services = nss, pam/services = nss, pam, ssh/" /etc/sssd/sssd.conf.tmp
        sed -i "s/^use_fully_qualified_names = True/use_fully_qualified_names = False/" /etc/sssd/sssd.conf.tmp
        sed -i "s/^fallback_homedir = \/home\/%u@%d/fallback_homedir = \/home\/%u/" /etc/sssd/sssd.conf.tmp
        echo "session    required    pam_mkhomedir.so    skel=/etc/skel/    umask=0022" >> /etc/pam.d/common-account
    fi

    mv -f /etc/sssd/sssd.conf.tmp /etc/sssd/sssd.conf; chmod 400 /etc/sssd/sssd.conf
    echo "%${adGroupName}@${directoryDNSName}  ALL=(ALL:ALL) ALL" >> /etc/sudoers
    
    if [[ ${OS_ID} == "ubuntu" ]]
    then
        systemctl restart realmd sssd
        systemctl enable realmd sssd
    else
        service sssd start
    fi

    # If 'autoadjoin' tag with 'false' value is found then do not reboot instance - instance owner will be responsible to reboot the instance.
    autoadjoinTagValue=$($awsbin --region ${awsRegion} ec2 describe-tags --filters "Name=resource-type,Values=instance" "Name=key,Values=autoadjoin" "Name=resource-id,Values=${instanceId}" --query "Tags[].Value" --output text)
    if [[ -n ${autoadjoinTagValue} && ${autoadjoinTagValue} == "false" ]]
    then
        logger "ADJoin: Found autoadjoin tag with value of false. Not Rebooting."
    else
        logger "ADJoin: Rebooting instance via SSM exit."
        exit 194
    fi
}